#define _RETAIL
#define GAME_DLL
#define WIN32_LEAN_AND_MEAN
#include <GMLuaModule.h>
#include "gm_botcontroller.h"
#include <engine/iserverplugin.h>
#include <eiface.h>
#include <iplayerinfo.h>
class ISave;
class IRestore;
class CTakeDamageInfo;
class touchlink_t;
class groundlink_t;
class notify_system_event_t;

IVEngineServer	*engine			= NULL;
IBotManager		*botmanager		= NULL;

#include <variant_t.h>
#include <predictioncopy.h>
#include <entitylist.h>
#include <baseentity.h>

#define META_BOT "Bot"
#define TYPE_BOT 7000

#define GETBOTFROMLUA() ILuaInterface *gLua = Lua(); \
		if (gLua->GetType(1) != TYPE_BOT) gLua->TypeError(META_BOT, 1); \
		edict_t *bot = (edict_t *)gLua->GetUserData(1); \
		if (!bot || bot->IsFree()) gLua->Error("Invalid bot!\n"); \
		IBotController *controller = botmanager->GetBotController( bot ); \
		if (!controller) gLua->Error("Invalid bot controller!\n");

LUA_FUNCTION(CreateBot)
{
	ILuaInterface *gLua = Lua();

	gLua->CheckType(1, GLua::TYPE_STRING);

	edict_t *bot = botmanager->CreateBot( gLua->GetString(1) );

	if(!bot)
	{
		gLua->Error("Unable to create bot!\n");
		return 0;
	}

	ILuaObject* Bot = gLua->GetMetaTable(META_BOT, TYPE_BOT);
	gLua->PushUserData(Bot, (void *)bot);
	Bot->UnReference();

	IServerUnknown *pUnkn = bot->GetUnknown();
	CBaseEntity *pEnt = pUnkn->GetBaseEntity();
	const CBaseHandle *pRef = &pEnt->GetRefEHandle();

	ILuaObject* Entity = gLua->GetMetaTable("Entity", GLua::TYPE_ENTITY);
	gLua->PushUserData(Entity, (void *)pRef);
	Entity->UnReference();

	return 2;
}

LUA_FUNCTION(BotPostSent)
{
	GETBOTFROMLUA();

	controller->PostClientMessagesSent();

	return 0;
}

LUA_FUNCTION(BotSetConVar)
{
	GETBOTFROMLUA();

	gLua->CheckType( 2, GLua::TYPE_STRING );
	gLua->CheckType( 3, GLua::TYPE_STRING );

	engine->SetFakeClientConVarValue( bot, gLua->GetString(2), gLua->GetString(3) );

	return 0;
}

LUA_FUNCTION(BotEFlagSet)
{
	GETBOTFROMLUA();

	gLua->CheckType(2, GLua::TYPE_NUMBER);

	bool bSet = controller->IsEFlagSet( gLua->GetInteger(2) );

	gLua->Push(bSet);

	return 1;
}

LUA_FUNCTION(BotSelect)
{
	GETBOTFROMLUA()

	gLua->CheckType(2, GLua::TYPE_STRING);

	controller->SetActiveWeapon( gLua->GetString(2) );

	return 0;
}

LUA_FUNCTION(BotStrip)
{
	GETBOTFROMLUA()

	controller->RemoveAllItems( true );

	return 0;
}

// this doesn't seem to actually do anything
LUA_FUNCTION(BotExec)
{
	GETBOTFROMLUA()
	
	gLua->CheckType(2, GLua::TYPE_STRING);

	engine->ClientCommand( bot, gLua->GetString( 2 ) );

	return 0;
}

// ( [forward], [side], [up], [viewangles], [buttons], [impulse], [weapon select], [weapon subtype] )
LUA_FUNCTION(BotMove)
{
	GETBOTFROMLUA()
	
	QAngle angle(0,0,0);
	if(gLua->GetType(5) == GLua::TYPE_ANGLE)
	{
		angle = *(QAngle *)gLua->GetUserData(5);
	}

	CBotCmd cmd;
	cmd.forwardmove		= gLua->GetInteger( 2 );
	cmd.sidemove		= gLua->GetInteger( 3 );
	cmd.upmove			= gLua->GetInteger( 4 );
	cmd.viewangles		= angle;
	cmd.buttons			= gLua->GetInteger( 6 );
	cmd.impulse			= gLua->GetInteger( 7 );
	cmd.weaponselect	= gLua->GetInteger( 8 );
	cmd.weaponsubtype	= gLua->GetInteger( 9 );

	controller->RunPlayerMove(&cmd);

	return 0;
}

int Start(lua_State *L)
{
	ILuaInterface *gLua = Lua();

	CreateInterfaceFn engineFactory		= Sys_GetFactory( "engine.dll" );
	CreateInterfaceFn interfaceFactory	= Sys_GetFactory( "server.dll" );

	engine		= (IVEngineServer	*) engineFactory(		INTERFACEVERSION_VENGINESERVER,		NULL);
	botmanager	= (IBotManager		*) interfaceFactory(	INTERFACEVERSION_PLAYERBOTMANAGER,	NULL);

	if(!botmanager)
	{
		gLua->Error("Missing botmanager interface.\n");
		return 0;
	}

	ILuaObject *table = gLua->GetNewTable();
		table->SetMember("CreateBot", CreateBot);
	gLua->SetGlobal("bot", table);
	table->UnReference();

	table = gLua->GetMetaTable(META_BOT, TYPE_BOT);
		ILuaObject* __index = gLua->GetNewTable();
			__index->SetMember("Move", BotMove);
			__index->SetMember("Exec", BotExec);
			__index->SetMember("RemoveAllItems", BotStrip);
			__index->SetMember("SetActiveWeapon", BotSelect);
			__index->SetMember("IsEFlagSet", BotEFlagSet);
			__index->SetMember("PostClientMessagesSent", BotPostSent);
			__index->SetMember("SetConVar", BotSetConVar);
		table->SetMember("__index", __index);
		__index->UnReference();
	table->UnReference();

	return 0;
}

int Close(lua_State *L)
{
	return 0;
}